import json
import string
import random
from uuid import uuid4
from decimal import Decimal
import datetime as dt


def get_random_eventid():
    """Returns a 32 char random hash garenteeing lowercase letter at the begining"""
    return random.choice(string.ascii_lowercase) + uuid4().hex[1:]


def json_loads(x, **kwargs):
    """Loads json for DynamoDB (Numerical types go to Decimal)"""
    return json.loads(x, parse_int=Decimal, parse_float=Decimal, **kwargs)


def json_dumps(x, **kwargs):
    """Dumps json for DynamoDB (Numerical types go to Decimal) use indent=4 if required"""
    return json.dumps(x, cls=DecimalEncoder, separators=(',', ': '), sort_keys=True, **kwargs)


def string_from_datetime(d, sep=None):
    """string format from python datetime"""
    return d.replace(tzinfo=None).isoformat(sep=sep, timespec='milliseconds')


class DecimalEncoder(json.JSONEncoder):
    """ Makes type conversions to allow JSON serialisation using data types commonly used in the project
        Decimal -> float
        datetime -> ISO string
    """

    def default(self, o, markers=None):
        if isinstance(o, Decimal):
            return float(o)
        if isinstance(o, dt.datetime):
            return string_from_datetime(o)
        return json.JSONEncoder(self).default(o, markers)



import csv
import io
import math
import json
import boto3
from decimal import Decimal
from collections import defaultdict
from tlx.util.common import get_uuid


def _pull_values(item):
    return {k: _set_types(v) for k, v in item.items()}


_func_map = defaultdict(
    lambda: lambda x: x, {
        'M': _pull_values,
        'N': Decimal,
        'L': lambda values: [_set_types(d) for d in values]
    }
)


def _set_types(v):
    v_key = list(v.keys())[0]
    returned_value = list(v.values())[0]  # Always one from dump
    return _func_map[v_key](returned_value)


def batch_write(table, items):
    """
        Basic Usage:
        >>> from tlx.dynamodb.batch import batch_write, get_ddb_table
        >>> table = get_ddb_table('Table-Name')
        >>> with open('results-20180821-091839.json', 'r') as f:
        ...     items = json.load(f)
        >>> batch_write(table, items)
    """

    table = get_ddb_table(table)

    with table.batch_writer() as batch:
        for item in items:
            batch.put_item(
                Item=item,
            )


def batch_delete(table, item_keys):
    """
        item_keys must be a list of dictionary of the keys required by the Table.
        e.g for a single key table: [{'id': 001, 'id': 002, ...}]
        for tables with a sort key, that must be included.
    """

    table = get_ddb_table(table)

    with table.batch_writer() as batch:
        for item_key in item_keys:
            batch.delete_item(Key=item_key)


def load_scan_dump(dump_file, table=None):
    """
        Loads the results of a scan opperation into a table.

        Details:
        Takes the output of a `scan` operation such as: `aws dynamodb scan --table-name <TableName>`
        and writes to an existing table. Similar to the `aws dynamodb batch-write-item` command except:
            - No limit to amount of items in the upload (25 with awscli)
            - Take the output of a table scan, requiring no reformatting
    """

    table = get_ddb_table(table)
    items = json.load(dump_file)['Items']
    # TODO: Make this a generator (problem is testing it)
    batch_write(table, [_pull_values(item) for item in items])


def load_from_csv(csv_file, table):
    """ CSV must conform to the following format:
            first row:  Field names
            second row: Field types. One of: ['N', 'S']

        N.B Only works for flat data structures. i.e Maps/Lists/Sets are not supported
    """

    table = get_ddb_table(table)

    with io.open(csv_file, newline='') as csvfile:
        data = list(csv.reader(csvfile))

    field_names, types = data[0], data[1]

    # Throws KeyError if missing. Only string and number are supported for csv
    def _format_number(x):
        tmp = Decimal(x if x else 'Nan')
        if math.isnan(tmp) or math.isinf(tmp):
            # Remove Inf and Nan, DynamoDB does support them
            return None
        return tmp

    _this_func_map = {
        'N': _format_number,
        'S': lambda x: str(x),
    }

    # Decimal Conversion if string field
    try:
        ddata = [  # Not a generator for testing
            {
                k1: v1
                for k1, v1 in (
                    (k, _this_func_map[t](v))
                    for k, t, v in zip(field_names, types, d)
                )
                if v1
            }
            for d in data[2:]
        ]
    except KeyError:
        raise Exception("load_from_csv only supports Dynamo Types {}".format(list(_this_func_map)))

    batch_write(table, ddata)


def load_json_dump(file_name, table_name, primary_key=False):
    """ Loads a file consisting of newline seperated items, in which each item is
        a json row object (Such as a BigQuery dump).

        If `primary_key` is provided the field is added to each item with a unique id as the sole partition key.
        If not provided, the input data must contain the keys of the dynamodb.
    """
    table = get_ddb_table(table_name)

    with open(file_name, 'r') as f:
        items = [
            json.loads(line.replace('\n', ''), parse_int=Decimal, parse_float=Decimal)
            for line in f
        ]

    if primary_key:
        for i in items:
            i[primary_key] = get_uuid()

    batch_write(table, items)


def get_ddb_table(table):
    """Takes either a string or an existing boto3 Table object.  If neither raises Exception"""

    try:  # test is boto3 Table object
        _ = table.name  # noqa: F841
    except AttributeError:
        if isinstance(table, str):
            dynamodb = boto3.resource('dynamodb')
            table = dynamodb.Table(table)
        else:
            raise Exception(("Table must be either a boto3 Table object"
                             "or a string not: {}").format(type(table)))
    return table



def paginate(method, **kwargs):
    """ Automatically paginates through result lists that yeild a 'nextToken' field
        e.g Get all Log groups rather than just the first 50
            >>> [lg['logGroupName'] for lg in paginate(logs.describe_log_groups)]
    """

    client = method.__self__
    paginator = client.get_paginator(method.__name__)
    for page in paginator.paginate(**kwargs).result_key_iters():
        try:
            for result in page:
                yield result
        except TypeError:
            pass
            # Dont know what the issue is here
            # but we get it sometimes with this:
            # ddb = boto3.client('dynamodb')
            # kwargs = {
            #     "TableName": 'TableName',
            #     "AttributesToGet": ('UserID',),
            # }
            # [x for x in tlx.util.paginate(ddb.scan, **kwargs)]



import sys
import logging
import functools
from tlx.dynamodb.aux import json_dumps
logger = logging.getLogger()
logger.setLevel(logging.INFO)


def proxy_response_handler(func=None, running_local=False, quiet=True):
    """ A Decorator for lambda functions. The function to be decorated by have two positional arguments
        (event, context) as any lambda handler would.  Decorating your handler allows you to write idomatic python
        using returns and raising exception.  This handler catches and formats them as proxy response objects
        suitable for APIG.

        This function does the following:
            - prints the received `event` to the debug logs
            - Sets the generic error response
            - Ensures raised exceptions are caught and formatted correctly
            - Unforseen errors are NOT propogated back to the user
            - Forseen exceptions ARE propogated back to the user in the `response.message` field
                (Developers should raise APIGException if they want the message to return)

        The decorated function should return data that can be converted to JSON.  This can be a list, dict, string,
        number or boolean.  It should raise a APIGException if the user wants to return the error message and modify
        the return code.  Otherwise all other Exceptions are returned as 500 response codes without a detailed error
        message.

        Set running_local=True if we're running locally (eg. we're not imported/running on AWS Lambda),
        and Python stack traces will show when exceptions are raised.
        Example usage: @proxy_response_handler(running_local=__name__=="__main__")

        Set quiet=True (default) to suppress all error output including stack traces (eg. for Prod deployments)
    """
    if not func:
        return functools.partial(proxy_response_handler, running_local=running_local, quiet=quiet)

    @functools.wraps(func)
    def wrapper(*axgs):
        # Setup default response
        response = {
            "statusCode": 500,
            "body": {
                "message": "Error",
                "response": {},
            },
        }

        def setup_error_response(msg, code=None):
            logger.error(msg)
            response["body"]["message"] = msg
            if code:
                response["statusCode"] = code

        # If the event is not from the apigateway a KeyError may be raised
        # and intentionally not caught
        event, context = axgs[0], axgs[1]
        logger.info('event: {}'.format(json_dumps(event)))
        logger.debug("Received '{resource}' request with params: {queryStringParameters} and body: {body}".format(**event))

        try:  # to get successfull execution
            response["body"]["response"] = func(event, context)
            response["body"]["message"] = "Success"
            response["statusCode"] = 200

        # if not, format appropriately for proxy integration
        except APIGException as e:
            setup_error_response("Error: {e}".format(e=e), e.code)
        except Exception as e:  # Unforseen Exception arose
            if quiet:
                pass  # Returns generic error response for production deployment
            elif running_local:
                sys.stderr.write("Local run detected, showing Python Exception stack trace:\n")
                raise e
            else:
                setup_error_response("Error: {e}".format(e=e))  # For remote testing

        # Final preparation for http reponse
        if response["body"]["response"] is None:
            del response["body"]["response"]
        response["body"] = json_dumps(response["body"])
        logger.info("Returning repsonse: {response}".format(**locals()))
        return response
    return wrapper


def require_valid_inputs(supplied, required):
    """ Returns None if `supplied` is a superset of `required`.  Raises `APIGException` with
        error code 400 if not.
        Bothe params must be an iterable.
    """

    try:
        missing_parameters = set(required).difference(supplied)
    except TypeError:
        # supplied is not itterable
        missing_parameters = required

    if missing_parameters:
        msg = "Invalid input parameters: {missing_parameters}".format(**locals())
        raise APIGException(msg, code=400)


class APIGException(Exception):
    """
        Differentiate these exceptions from general ones so that we can return the exception message.
        (We don't want to return general exception messages)

        e.g to return a HTTP 402 with a custom message, raise an exception like so:
        `raise APIGException('Payment required', code=402)`
    """
    def __init__(self, message, code=500):
        self.code = code
        Exception.__init__(self, message)
